@@ -18,6 +18,7 @@
 
 import fileinput
 import itertools
+import json
 import logging
 import os
 import pathlib
@@ -25,11 +26,12 @@ import shlex
 import socket
 import subprocess
 import time
-import uuid
 import yaml
+import requests
+import medusa.utils
 
 from subprocess import PIPE
-
+from retrying import retry
 from cassandra.cluster import Cluster, ExecutionProfile
 from cassandra.policies import WhiteListRoundRobinPolicy
 from cassandra.auth import PlainTextAuthProvider
@@ -92,7 +94,6 @@ class CqlSessionProvider(object):
         if retry:
             max_retries = 5
             attempts = 0
-            delay = 5
 
             while attempts < max_retries:
                 try:
@@ -100,6 +101,7 @@ class CqlSessionProvider(object):
                     return CqlSession(session, self._cassandra_config.resolve_ip_addresses)
                 except Exception as e:
                     logging.debug('Failed to create session', exc_info=e)
+                delay = 5 * (2 ** (attempts + 1))
                 time.sleep(delay)
                 attempts = attempts + 1
             raise Exception('Could not establish CQL session after {attempts}'.format(attempts=attempts))
@@ -229,11 +231,12 @@ class CassandraConfigReader(object):
         config_file = pathlib.Path(cassandra_config or self.DEFAULT_CASSANDRA_CONFIG)
         if not config_file.is_file():
             raise RuntimeError('{} is not a file'.format(config_file))
-        self._config = yaml.load(config_file.open(), Loader=yaml.BaseLoader)
+        with open(config_file, 'r') as f:
+            self._config = yaml.load(f, Loader=yaml.BaseLoader)
 
     @property
     def root(self):
-        data_file_directories = self._config.get('data_file_directories')
+        data_file_directories = self._config.get('data_file_directories', ['/var/lib/cassandra/data'])
         if not data_file_directories:
             raise RuntimeError('data_file_directories must be properly configured')
         if len(data_file_directories) > 1:
@@ -242,14 +245,14 @@ class CassandraConfigReader(object):
 
     @property
     def commitlog_directory(self):
-        commitlog_directory = self._config.get('commitlog_directory')
+        commitlog_directory = self._config.get('commitlog_directory', '/var/lib/cassandra/commitlog')
         if not commitlog_directory:
             raise RuntimeError('commitlog_directory must be properly configured')
         return pathlib.Path(commitlog_directory)
 
     @property
     def saved_caches_directory(self):
-        saved_caches_directory = self._config.get('saved_caches_directory')
+        saved_caches_directory = self._config.get('saved_caches_directory', '/var/lib/cassandra/saved_caches')
         if not saved_caches_directory:
             raise RuntimeError('saved_caches_directory must be properly configured')
         return pathlib.Path(saved_caches_directory)
@@ -288,15 +291,18 @@ class CassandraConfigReader(object):
         seeds = list()
         if 'seed_provider' in self._config:
             if self._config['seed_provider']:
-                return self._config.get('seed_provider')[0]['parameters'][0]['seeds'].replace(' ', '').split(',')
+                if self._config['seed_provider'][0]['class_name'].endswith('SimpleSeedProvider'):
+                    return self._config.get('seed_provider')[0]['parameters'][0]['seeds'].replace(' ', '').split(',')
         return seeds
 
 
 class Cassandra(object):
 
     SNAPSHOT_PATTERN = '*/*/snapshots/{}'
+    SNAPSHOT_PREFIX = 'medusa-'
 
-    def __init__(self, cassandra_config, contact_point=None):
+    def __init__(self, config, contact_point=None):
+        cassandra_config = config.cassandra
         self._start_cmd = shlex.split(cassandra_config.start_cmd)
         self._stop_cmd = shlex.split(cassandra_config.stop_cmd)
         self._is_ccm = int(shlex.split(cassandra_config.is_ccm)[0])
@@ -319,6 +325,9 @@ class Cassandra(object):
         self._rpc_port = config_reader.rpc_port
         self.seeds = config_reader.seeds
 
+        self.grpc_config = config.grpc
+        self.kubernetes_config = config.kubernetes
+
     def _has_systemd(self):
         try:
             result = subprocess.run(['systemctl', '--version'], stdout=PIPE, stderr=PIPE)
@@ -404,34 +413,68 @@ class Cassandra(object):
         def __repr__(self):
             return '{}<{}>'.format(self.__class__.__qualname__, self._tag)
 
-    def create_snapshot(self):
-        tag = 'medusa-{}'.format(uuid.uuid4())
-        cmd = self._nodetool.nodetool + ['snapshot', '-t', tag]
+    def create_snapshot(self, backup_name):
+        cmd = self.create_snapshot_command(backup_name)
+        tag = "{}{}".format(self.SNAPSHOT_PREFIX, backup_name)
+        if not self.snapshot_exists(tag):
 
-        if self._is_ccm == 1:
-            os.popen('ccm node1 nodetool \"snapshot -t {}\"'.format(tag)).read()
-        else:
-            logging.debug('Executing: {}'.format(' '.join(cmd)))
-            subprocess.check_call(cmd, stdout=subprocess.DEVNULL, universal_newlines=True)
+            # TODO introduce abstraction layer/interface for invoking Cassandra
+            # Eventually I think we will want to introduce an abstraction layer for Cassandra's
+            # API that Medusa requires. There should be an implementation for using nodetool,
+            # one for Jolokia, and a 3rd for the management sidecard used by Cass Operator.
+            if medusa.utils.evaluate_boolean(self.kubernetes_config.enabled):
+                data = {
+                    "type": "exec",
+                    "mbean": "org.apache.cassandra.db:type=StorageService",
+                    "operation": "takeSnapshot(java.lang.String,java.util.Map,[Ljava.lang.String;)",
+                    "arguments": [tag, {}, []]
+                }
+                response = self.__do_post(data)
+                if response["status"] != 200:
+                    raise Exception("failed to create snapshot: {}".format(response["error"]))
+            else:
+                if self._is_ccm == 1:
+                    os.popen(cmd).read()
+                else:
+                    logging.debug('Executing: {}'.format(' '.join(cmd)))
+                    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, universal_newlines=True)
 
         return Cassandra.Snapshot(self, tag)
 
     def delete_snapshot(self, tag):
-        cmd = self._nodetool.nodetool + ['clearsnapshot', '-t', tag]
+        cmd = self.delete_snapshot_command(tag)
+        if self.snapshot_exists(tag):
 
-        if self._is_ccm == 1:
-            os.popen('ccm node1 nodetool \"clearsnapshot -t {}\"'.format(tag)).read()
-        else:
-            logging.debug('Executing: {}'.format(' '.join(cmd)))
-            try:
-                output = subprocess.check_output(cmd, universal_newlines=True)
-                logging.debug('nodetool output: %s', output)
-            except subprocess.CalledProcessError as e:
-                logging.debug('nodetool resulted in error: %s', e.output)
-                logging.warning(
-                    'Medusa may have failed at cleaning up snapshot %s. '
-                    'Check if the snapshot exists and clear it manually by running: %s',
-                    tag, ' '.join(cmd))
+            if medusa.utils.evaluate_boolean(self.kubernetes_config.enabled):
+                data = {
+                    "type": "exec",
+                    "mbean": "org.apache.cassandra.db:type=StorageService",
+                    "operation": "clearSnapshot",
+                    "arguments": [tag, []]
+                }
+                response = self.__do_post(data)
+                if response["status"] != 200:
+                    raise Exception("failed to delete snapshot: {}".format(response["error"]))
+            else:
+                if self._is_ccm == 1:
+                    os.popen(cmd).read()
+                else:
+                    logging.debug('Executing: {}'.format(' '.join(cmd)))
+                    try:
+                        output = subprocess.check_output(cmd, universal_newlines=True)
+                        logging.debug('nodetool output: {}'.format(output))
+                    except subprocess.CalledProcessError as e:
+                        logging.debug('nodetool resulted in error: {}'.format(e.output))
+                        logging.warning(
+                            'Medusa may have failed at cleaning up snapshot {}. '
+                            'Check if the snapshot exists and clear it manually '
+                            'by running: {}'.format(tag, ' '.join(cmd)))
+
+    def __do_post(self, data):
+        json_data = json.dumps(data)
+        response = requests.post(self.kubernetes_config.cassandra_url, data=json_data)
+
+        return json.loads(response.text)
 
     def list_snapshotnames(self):
         return {
@@ -452,6 +495,29 @@ class Cassandra(object):
                 return True
         return False
 
+    def create_snapshot_command(self, backup_name):
+        """
+        :param backup_name: string name of the medusa backup
+        :return: Array representation of a command to create a snapshot
+        """
+        tag = '{}{}'.format(self.SNAPSHOT_PREFIX, backup_name)
+        if self._is_ccm == 1:
+            cmd = 'ccm node1 nodetool \"snapshot -t {}\"'.format(tag)
+        else:
+            cmd = self._nodetool.nodetool + ['snapshot', '-t', tag]
+        return cmd
+
+    def delete_snapshot_command(self, tag):
+        """
+        :param tag: string snapshot name
+        :return: Array repesentation of a command to delete a snapshot
+        """
+        if self._is_ccm == 1:
+            cmd = 'ccm node1 nodetool \"clearsnapshot -t {}\"'.format(tag)
+        else:
+            cmd = self._nodetool.nodetool + ['clearsnapshot', '-t', tag]
+        return cmd
+
     def _columnfamily_path(self, keyspace_name, columnfamily_name, cf_id):
         root = pathlib.Path(self._root)
         keyspace_path = root / keyspace_name / columnfamily_name
@@ -540,45 +606,27 @@ class Cassandra(object):
                 cassandra_yaml.write('\nauto_bootstrap: false')
 
 
-def wait_for_node_to_come_up(config, host, retries=10, delay=6):
-    logging.info('Waiting for Cassandra to come up on %s', host)
-    check_function = is_node_up
-    if not _wait_for_node(config, host, check_function, retries, delay):
-        raise CassandraNodeNotUpError(host, retries)
-    logging.info('Cassandra is up on %s', host)
+@retry(stop_max_attempt_number=7, wait_exponential_multiplier=5000, wait_exponential_max=120000)
+def wait_for_node_to_come_up(config, host):
+    logging.info('Waiting for Cassandra to come up on {}'.format(host))
+
+    if not is_node_up(config, host):
+        raise CassandraNodeNotUpError(host)
+    else:
+        logging.info('Cassandra is up on {}'.format(host))
+        return True
 
 
-def wait_for_node_to_go_down(config, host, retries=10, delay=6):
-    logging.info('Waiting for Cassandra to go down on %s', host)
+@retry(stop_max_attempt_number=7, wait_exponential_multiplier=5000, wait_exponential_max=120000)
+def wait_for_node_to_go_down(config, host):
+    logging.info('Waiting for Cassandra to go down on {}'.format(host))
 
-    def check_function(config, host):
-        return not is_node_up(config, host)
-
-    if not _wait_for_node(config, host, check_function, retries, delay):
+    if is_node_up(config, host):
         # TODO can do a kill here
-        raise CassandraNodeNotDownError(host, retries)
-    logging.info('Cassandra is down %s', host)
-
-
-def _wait_for_node(config, host, check_function, retries, delay):
-    """
-        Polls the node until the check_function check passes.
-
-        :param host: The target host on which to run the check
-        :param check_function: called to determine node state
-        :param retries: The number of times to retry the health check. Defaults to 10
-        :param delay: A delay in seconds to wait before polling again. Defaults to 6 seconds.
-        :return: True when the node is determined to be up. If the retries are exhausted, returns False
-    """
-    attempts = 0
-    while attempts < retries:
-        if check_function(config, host):
-            return True
-        else:
-            time.sleep(delay)
-            attempts = attempts + 1
-
-    return False
+        raise CassandraNodeNotDownError(host)
+    else:
+        logging.info('Cassandra is down {}'.format(host))
+        return True
 
 
 def is_node_up(config, host):
@@ -604,40 +652,58 @@ def is_node_up(config, host):
         else:
             return is_ccm_up(args, 'statusbinary')
     else:
-        cassandra = Cassandra(config.cassandra)
+        cassandra = Cassandra(config)
         native_port = cassandra.native_port
         rpc_port = cassandra.rpc_port
-        nc_timeout = 10
-        args = ['timeout', str(nc_timeout), 'nc', '-zv', host]
         if health_check == 'thrift':
-            return is_cassandra_up(args, rpc_port)
+            return is_cassandra_up(host, rpc_port)
         elif health_check == 'all':
-            return is_cassandra_up(list(args), rpc_port) and is_cassandra_up(list(args), native_port)
+            return is_cassandra_up(host, rpc_port) and is_cassandra_up(host, native_port)
         else:
             # cql only
-            return is_cassandra_up(args, native_port)
+            return is_cassandra_up(host, native_port)
 
 
 def is_ccm_up(args, nodetool_command):
     try:
         args.append(nodetool_command)
         output = subprocess.check_output(args, stderr=subprocess.STDOUT, universal_newlines=True)
-        return output.find('running') >= 0
+        # ccm returns either 'running' or 'not running'.
+        # Testing on finding 'running' is not enough!
+        if output.find('not running') == -1 and output.find('running') >= 0:
+            logging.debug('CCM native transport is now up')
+            return True
+        else:
+            logging.debug('CCM native transport is not up yet')
+            return False
     except subprocess.CalledProcessError as e:
-        # logging.debug('The native transport is not up yet %s', logging_suffix)
-        logging.debug('The native transport is not up yet', exc_info=e)
+        logging.debug('CCM native transport is not up yet', exc_info=e)
         return False
 
 
-def is_cassandra_up(args, port):
+def is_open(host, port):
+    timeout = 3
+    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+    s.settimeout(timeout)
+    is_accessible = False
     try:
-        args.append(port)
-        subprocess.check_call(args)
-        # check_call returns if call succeeded (returncode is 0). If not exception is triggered
+        s.connect((host, port))
+        s.shutdown(socket.SHUT_RDWR)
+        is_accessible = True
+    except socket.error as e:
+        logging.debug('Port {} close on host {}'.format(port, host), exc_info=e)
+        is_accessible = False
+    finally:
+        s.close()
+        return is_accessible
+
+
+@retry(stop_max_attempt_number=5, wait_exponential_multiplier=5000, wait_exponential_max=120000)
+def is_cassandra_up(host, port):
+    if is_open(host, port):
         return True
-    except subprocess.CalledProcessError:
-        logging.debug('The node is not up yet')
-        return False
+    else:
+        logging.debug('The node {} is not up yet...'.format(host))
 
 
 class CassandraNodeNotUpError(Exception):
@@ -650,8 +716,8 @@ class CassandraNodeNotUpError(Exception):
         attempts -- the number of times the check was performed
     """
 
-    def __init(self, host, attempts):
-        msg = 'Could not verify that Cassandra is up on {host} after {attempts}'.format(host=host, attempts=attempts)
+    def __init(self, host):
+        msg = 'Cassandra node {} is still down...'.format(host)
         super(CassandraNodeNotUpError, self).__init__(msg)
 
 
@@ -660,6 +726,6 @@ class CassandraNodeNotDownError(Exception):
     Raised when we give up waiting on a node to go down, meaning nodetool statusbinary and/or statusthrift keep
     reporting open ports.
     """
-    def __init(self, host, attempts):
-        msg = 'Could not verify that Cassandra is down on {host} after {attempts}'.format(host=host, attempts=attempts)
+    def __init(self, host):
+        msg = 'Cassandra node {} is still up...'.format(host)
         super(CassandraNodeNotDownError, self).__init__(msg)
