@@ -30,6 +30,9 @@ from medusa.storage.abstract_storage import AbstractStorage
 from medusa.storage.google_cloud_storage.gsutil import GSUtil
 
 
+GSUTIL_MAX_FILES_PER_CHUNK = 64
+
+
 class GoogleStorage(AbstractStorage):
 
     def connect_storage(self):
@@ -74,7 +77,11 @@ class GoogleStorage(AbstractStorage):
         # so the final upload destination is `dest / parent / object`
         # this is needed to handle things like secondary indices that live in hidden folders within table folders
         new_dest = '{}/{}'.format(old_dest, parent) if parent.startswith('.') else old_dest
-        return gsutil.cp(srcs=src_paths, dst="gs://{}/{}".format(self.bucket.name, new_dest))
+        return gsutil.cp(
+            srcs=src_paths,
+            dst="gs://{}/{}".format(self.bucket.name, new_dest),
+            parallel_process_count=self.config.concurrent_transfers
+        )
 
     def download_blobs(self, srcs, dest):
 
